"""
The :code:`sensor` module bundles all ARTS functionality related to sensors
and provides a set of predefined sensor objects.
"""
from artssat.sensor.sensor import ICI, MWI, ActiveSensor, PassiveSensor, CloudSat





