"""
Handling of PSD data
====================

The :code:`artssat.scattering.psd.data` subpackage provides the
:class:`PSDData` class for the handling of numeric PSD data. Numeric PSD data
here refers to a PSD that is not given in a parametrized form but instead by
an array of particle density values over a given size grid.
"""
