"""
ARTS PSD interface
==================

The ARTS interface is implemented by the :class:`ArtsPSD` class, which
implements common functionality for the handling of PSDs in ARTS and defines
the interaction with the :class:`ArtsSimulation` and in particular
the :class:`ScatteringSpecies` instance in which the PSD is used.
"""

